#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.extra.mdashboard
      @file: dashboard_item.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Callable

from hspylib.modules.cli.icons.font_awesome.awesome import Awesome


class DashboardItem:
    """TODO"""

    def __init__(
        self,
        icon: Awesome = None,
        tooltip: str = None,
        on_trigger: Callable = None):
        self.icon = icon
        self.tooltip = tooltip
        self.on_trigger = on_trigger

    def __str__(self):
        return f"{self.icon} \"{self.tooltip}\" -> {self.on_trigger}"

    def __repr__(self):
        return str(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db
      @file: sqlite_configuration.py
   @created: Thu, 03 Nov 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Optional

from hspylib.core.datasource.db_configuration import DBConfiguration


class SQLiteConfiguration(DBConfiguration):
    """TODO"""

    def __init__(
        self,
        resource_dir: str,
        filename: Optional[str] = None,
        profile: Optional[str] = None):

        super().__init__(resource_dir, filename, profile)
        self._db_file = self['datasource.db.file']

    @property
    def db_file(self) -> str:
        return self._db_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.eventbus
      @file: eventbus.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from typing import Any, Callable, List

from hspylib.core.exception.exceptions import HSBaseException
from hspylib.core.preconditions import check_argument
from hspylib.modules.eventbus.event import Event


def subscribe(**kwargs) -> Callable:
    """Method decorator to subscribe to a given bus event."""

    def inner(func) -> None:
        check_argument(func.__code__.co_argcount >= 1, 'Subscriber callbacks require at least one parameter.')
        missing = next((p for p in ['bus', 'event'] if p not in kwargs.keys()), None)
        check_argument(missing is None, f"Missing required parameter: '{missing}: str'.")
        EventBus.get(str(kwargs['bus'])).subscribe(str(kwargs['event']), func)

    return inner


def emit(bus_name: str, event_name: str, **kwargs) -> None:
    """Emit an event to the specified bus."""
    EventBus.get(bus_name).emit(event_name, **kwargs)


class EventBus:
    """Provide an eventbus pattern for events and subscribers."""

    _buses = {}
    _subscribers = {}
    _events: List[Event] = []

    @classmethod
    def get(cls, bus_name: str) -> 'EventBus':
        """Return the bus instance referred to the specified bus name."""
        if bus_name in cls._buses:
            return cls._buses[bus_name]
        bus_instance = EventBus(bus_name)
        cls._buses[bus_name] = bus_instance
        return bus_instance

    @classmethod
    def _get_subscriber(cls, bus_name: str, event_name: str) -> Any:
        """Return the subscriber of the referred bus name and event name."""
        cache_key = f'{bus_name}.{event_name}'
        if cache_key in cls._subscribers:
            return cls._subscribers[cache_key]
        subscriber = {'callbacks': []}
        cls._subscribers[cache_key] = subscriber
        return subscriber

    def __init__(self, name: str):
        self.name = name

    def subscribe(self, event_name: str, cb_event_handler: Callable) -> None:
        """Subscribe to the specified event bus."""
        subscriber = self._get_subscriber(self.name, event_name)
        subscriber['callbacks'].append(cb_event_handler)

    def emit(self, event_name: str, **kwargs) -> None:
        """Emit an event to this bus."""
        self._events.append(Event(event_name, **kwargs))
        while len(self._events) > 0:
            event = self._events.pop()
            cache_key = f"{self.name}.{event.name}"
            subscribers = self._subscribers[cache_key] if cache_key in self._subscribers else None
            if subscribers and len(subscribers['callbacks']) > 0:
                for callback in subscribers['callbacks']:
                    try:
                        callback(event)
                    except Exception as err:
                        raise HSBaseException(
                            f"{self.__class__.__name__}::emit Callback invocation failed - {str(err)}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db.firebase
      @file: firebase_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import json
import logging as log
from abc import abstractmethod
from typing import Iterable, List, Optional, Set, TypeVar

from requests.exceptions import HTTPError

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.crud_repository import CrudRepository
from hspylib.core.datasource.firebase.firebase_configuration import FirebaseConfiguration
from hspylib.core.datasource.identity import Identity
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.namespace import Namespace
from hspylib.modules.fetch.fetch import delete, get, put

T = TypeVar('T', bound=CrudEntity)


class FirebaseRepository(CrudRepository[T]):
    """Implementation of a data access layer for a Firebase persistence store."""

    def __init__(self, config: FirebaseConfiguration):
        self._payload = None
        self._config = config

    def __str__(self) -> str:
        return str(self._config) + f"TABLE_NAME={self.table_name() or ''}"

    def __repr__(self):
        return str(self)

    @property
    def config(self) -> FirebaseConfiguration:
        return self._config

    @property
    def info(self) -> str:
        return f"{self.config}"

    @property
    def hostname(self) -> str:
        return self._config.hostname

    @property
    def port(self) -> int:
        return self._config.port

    @property
    def base_url(self) -> str:
        return self._config.base_url

    @property
    def database(self) -> str:
        """Return the database name."""
        return self._config.database

    def delete(self, entity: T) -> None:
        """Deletes the given entity from the Firebase store"""
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        """Deletes the given entity by it's Id from the Firebase store"""
        ids = '.'.join(entity_id.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        log.debug('Deleting firebase entry: \n\t|-Id=%s\n\t|-From %s', entity_id, url)
        response = delete(url)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to delete from={url}')

    def delete_all(self, entities: List[T]) -> None:
        """Delete from Firebase all entries provided.
        TODO Check if there is a better ways of doing it.
        """
        list(map(self.delete, entities))

    def save(self, entity: T, exclude_update: Iterable[str] = None) -> None:
        """Saves the given entity at the Firebase store"""
        ids = '.'.join(entity.identity.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        payload = entity.as_json()
        log.debug("Saving firebase entry: \n\t|-%s \n\t|-Into %s", entity, url)
        response = put(url, payload)
        check_not_none(response, "Response is none")
        if response.status_code not in [HttpCode.OK, HttpCode.ACCEPTED]:
            raise HTTPError(f'{response.status_code} - Unable to put into={url} with json_string={payload}')

    def save_all(self, entities: List[T], exclude_update: Iterable[str] = None) -> None:
        """Save from Firebase all entries provided.
        TODO Check if there is a better ways of doing it.
        """
        list(map(self.save, entities))

    def find_all(
        self,
        fields: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500, offset: int = 0) -> List[T]:
        """Return filtered entries from the Firebase store"""

        # orders = "orderBy=" + (','.join([f'"{o}"' for o in order_bys]) if order_bys else '"$key"')
        # limits = f"startAt={offset}&endAt={offset + limit}"
        url = f'{self._config.url(self.table_name())}.json'
        log.debug('Fetching firebase entries: \n\t|-From %s', url)
        response = get(url)
        check_not_none(response, "Response is none")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to get from={url}')

        if response.body and response.body != 'null':
            return self.to_entity_list(response.body, filters)[offset:offset + limit]

        return []

    def find_by_id(self, entity_id: Identity, fields: Set[str] = None) -> Optional[T]:
        """Return the entry specified by ID from the Firebase store, None if no such entry is found."""
        ids = '.'.join(entity_id.values)
        url = f'{self._config.url(self.table_name())}/{ids}.json'
        log.debug('Fetching firebase entry: \n\t|-Id=%s\n\t|-From %s', entity_id, url)
        response = get(url)
        check_not_none(response, "Response is null")
        if response.status_code != HttpCode.OK:
            raise HTTPError(f'{response.status_code} - Unable to get from={url}')
        if response.body and response.body != 'null':
            return self.to_entity_type(json.loads(response.body))

        return None

    def exists_by_id(self, entity_id: Identity) -> bool:
        """Check if provided entity exists on the database.
        TODO Check if there is a better ways of doing it.
        """
        return self.find_by_id(entity_id) is not None

    @abstractmethod
    def table_name(self) -> str:
        """TODO"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: hspylib
   @package: hspylib.main.hspylib.core.tools
      @file: collection_filter.py
   @created: Wed, 8 Jun 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import get_args, Iterable, Iterator, Set, Tuple, TypeVar, Union

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.text_tools import quote

T = TypeVar('T')
FILTER_VALUE = TypeVar('FILTER_VALUE', int, str, bool, float)


class FilterCondition(Enumeration):
    """Collection filter conditions."""

    # @formatter:off
    LESS_THAN                   =      '<', Union[int, float]
    LESS_THAN_OR_EQUALS_TO      =     '<=', Union[int, float]
    GREATER_THAN                =      '>', Union[int, float]
    GREATER_THAN_OR_EQUALS_TO   =     '>=', Union[int, float]
    EQUALS_TO                   =     '==', Union[str, int, bool, float]
    DIFFERENT_FROM              =     '!=', Union[str, int, bool, float]
    CONTAINS                    =     'in', Union[str]
    DOES_NOT_CONTAIN            = 'not in', Union[str]
    IS                          =     '==', Union[bool]
    IS_NOT                      =     '!=', Union[bool]
    # @formatter:on

    def __str__(self):
        return self.name.lower().replace('_', ' ')

    def __repr__(self):
        return str(self)

    def matches(self, param_value: FILTER_VALUE, value: FILTER_VALUE) -> bool:
        """Whether this filter value matches the specified param."""
        operator = self.value[0]
        if self.name in ['CONTAINS', 'DOES_NOT_CONTAIN']:
            expression = f'{quote(value)} {operator} {quote(param_value)}'
        else:
            expression = f'{quote(param_value)} {operator} {quote(value)}'
        return self._allow_type(value) and eval(expression)  # pylint: disable=eval-used

    def _allow_type(self, value: FILTER_VALUE) -> bool:
        """Whether this filter condition allows the specified value type."""
        try:
            return isinstance(value, self.value[1])
        except TypeError:
            return isinstance(value, get_args(self.value[1]))


class ElementFilter:
    """Represent a single filter condition."""

    def __init__(
        self, name: str,
        el_name: str,
        condition: 'FilterCondition',
        el_value: FILTER_VALUE):
        self.name = name
        self.el_name = el_name
        self.condition = condition
        self.el_value = el_value

    def __str__(self):
        return f'{quote(self.el_name)} {self.condition} {quote(self.el_value)}'

    def __repr__(self):
        return str(self)

    def __key(self) -> Tuple[str, 'FilterCondition', FILTER_VALUE]:
        return self.el_name, self.condition, self.el_value

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other: 'ElementFilter') -> bool:
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return NotImplemented

    def matches(self, element: T) -> bool:
        """Whether this filter is True for the given element."""

        try:
            entry = None
            if isinstance(element, dict):
                entry = element
            elif hasattr(element, '__dict__'):
                entry = element.__dict__
            elif isinstance(element, tuple):
                entry = {k: v for k, v in element}
            return self.el_name in entry and self.condition.matches(entry[self.el_name], self.el_value)
        except (NameError, TypeError, AttributeError):
            return False


class CollectionFilter:
    """A collection of filters to be applied to a given iterable."""

    def __init__(self) -> None:
        self._filters: Set[ElementFilter] = set()

    def __str__(self) -> str:
        if len(self._filters) > 0:
            return ' '.join([
                f"{'AND ' if i > 0 else ''}{str(f)}" for i, f in enumerate(self._filters)
            ])
        return 'No filters applied'

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self) -> Iterator:
        return self._filters.__iter__()

    def __len__(self) -> int:
        return len(self._filters)

    def apply_filter(
        self,
        name: str,
        el_name: str,
        condition: 'FilterCondition',
        el_value: Union[int, str, bool, float]) -> None:
        """Apply the specified filter."""

        check_argument(not any(f.name == name for f in self._filters),
                       f'Filter {name} already exists!')
        f = ElementFilter(name, el_name, condition, el_value)
        self._filters.add(f)

    def clear(self) -> None:
        """Clear all filters."""
        self._filters.clear()

    def discard(self, name: str):
        """Discard the specified filter."""
        element = next((e for e in self._filters if e.name == name), None)
        self._filters.discard(element)

    def filter(self, data: Iterable[T]) -> Iterable[T]:
        """Filter the collection."""
        filtered: Iterable[T] = data.__class__.__call__()
        for element in data:
            if not self.should_filter(element):
                if hasattr(filtered, 'append'):
                    filtered.append(element)
                elif hasattr(filtered, 'add'):
                    filtered.add(element)
        return filtered

    def filter_inverse(self, data: Iterable[T]) -> Iterable[T]:
        """Inverse filter the collection."""
        filtered: Iterable[T] = data.__class__.__call__()
        for element in data:
            if self.should_filter(element):
                if hasattr(filtered, 'append'):
                    filtered.append(element)
                elif hasattr(filtered, 'add'):
                    filtered.add(element)
        return filtered

    def should_filter(self, data: T) -> bool:
        """Whether the specified data should be filtered, according to this filter collection."""
        return not all(f.matches(data) for f in self._filters)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.config
      @file: app_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
from textwrap import dedent
from typing import Any, Optional

from hspylib.core.config.properties import Properties
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import run_dir, str_to_bool


class AppConfigs(metaclass=Singleton):
    """Holds all of the application configurations (properties)"""

    INSTANCE = None

    DISPLAY_FORMAT = dedent("""
    AppConfigs
      |-Working dir = {}
      |-Resources dir = {}
      |-Properties:
       \\-{}
    """)

    def __init__(self, resource_dir: str, filename: Optional[str] = None, profile: Optional[str] = None):
        check_argument(os.path.exists(resource_dir),
                       "Unable to locate resources dir: {}", resource_dir)
        self._resource_dir = resource_dir
        self._properties = Properties(filename=filename, load_dir=resource_dir, profile=profile)
        log.info(self)

    def __str__(self) -> str:
        return '\n{}{}{}'.format(
            '-=' * 40,
            self.DISPLAY_FORMAT.format(
                str(run_dir()),
                str(self._resource_dir),
                str(self._properties).replace('\n', '\n   |-') if self._properties.size > 0 else ''
            ),
            '-=' * 40
        )

    def __repr__(self):
        return str(self)

    def __getitem__(self, item: str) -> Any:
        return self.get(item)

    def __len__(self) -> int:
        return self.size

    @property
    def resource_dir(self) -> Optional[str]:
        """Return the configured application resource dir"""
        return self._resource_dir

    @property
    def properties(self) -> Properties:
        """Return the application properties"""
        return self._properties

    @property
    def size(self) -> int:
        """Return the application properties"""
        return self._properties.size

    def get(self, property_name: str) -> Optional[str]:
        """Get the value, as a string, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name)

    def get_int(self, property_name: str) -> Optional[int]:
        """Get the value, as an integer, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, prop_type=int)

    def get_float(self, property_name: str) -> Optional[float]:
        """Get the value, as a float, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, prop_type=float)

    def get_bool(self, property_name: str) -> Optional[bool]:
        """Get the value, as a boolean, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, prop_type=str_to_bool)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.extra
      @file: minput.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import re
import time
from typing import List, Optional

from hspylib.core.exception.exceptions import InvalidInputError
from hspylib.core.tools.commons import syserr, sysout
from hspylib.core.tools.namespace import Namespace
from hspylib.core.tools.text_tools import snakecase
from hspylib.modules.cli.icons.font_awesome.form_icons import FormIcons
from hspylib.modules.cli.icons.font_awesome.nav_icons import NavIcons
from hspylib.modules.cli.keyboard import Keyboard
from hspylib.modules.cli.tui.extra.minput.form_builder import FormBuilder
from hspylib.modules.cli.tui.extra.minput.form_field import FormField
from hspylib.modules.cli.tui.extra.minput.input_type import InputType
from hspylib.modules.cli.tui.extra.minput.minput_utils import MInputUtils
from hspylib.modules.cli.vt100.vt_codes import vt_print
from hspylib.modules.cli.vt100.vt_colors import VtColors
from hspylib.modules.cli.vt100.vt_utils import get_cursor_position, prepare_render, restore_cursor, restore_terminal, \
    set_enable_echo


def minput(
    form_fields: List[FormField],
    title: str = 'Please fill all fields of the form fields below',
    prefix: str = None,
    title_color: VtColors = VtColors.ORANGE,
    nav_color: VtColors = VtColors.YELLOW) -> Optional[Namespace]:
    """
    TODO
    :param form_fields:
    :param title:
    :param prefix:
    :param title_color:
    :param nav_color:
    :return:
    """
    return MenuInput(form_fields).input(title, prefix, title_color, nav_color)


class MenuInput:
    """TODO"""

    SELECTED_BG = '%MOD(44)%'
    NAV_ICONS = NavIcons.compose(NavIcons.UP, NavIcons.DOWN)
    NAV_BAR = f"[Enter] Submit  [{NAV_ICONS}] Navigate  [{NavIcons.TAB}] Next  [Space] Toggle  [Esc] Quit %EL0%"

    @classmethod
    def builder(cls) -> FormBuilder:
        return FormBuilder()

    def __init__(self, all_fields: List[FormField]):
        self.all_fields = all_fields
        self.all_pos = [(0, 0) for _ in all_fields]
        self.cur_field = self.done = None
        self.cur_row = self.cur_col = self.tab_index = 0
        self.max_label_length = max(len(field.label) for field in all_fields)
        self.max_value_length = max(field.max_length for field in all_fields)
        self.max_detail_length = max(MInputUtils.detail_len(field) for field in all_fields)
        self.re_render = True

    def input(
        self,
        title: str,
        prefix: str,
        title_color: VtColors,
        nav_color: VtColors) -> Optional[Namespace]:
        """TODO"""

        ret_val = Keyboard.VK_NONE
        length = len(self.all_fields)

        if length == 0:
            return None

        prepare_render(title, title_color)

        # Wait for user interaction
        while not self.done and ret_val not in [Keyboard.VK_ENTER, Keyboard.VK_ESC]:
            # Menu Renderization
            if self.re_render:
                self._render(nav_color)

            # Navigation input
            ret_val = self._nav_input()

        restore_terminal()

        if ret_val == Keyboard.VK_ENTER:
            form_fields = Namespace('FormFields')
            for field in self.all_fields:
                att_name = f"{prefix or ''}{snakecase(field.label)}"
                form_fields.setattr(att_name, field.value)
            return form_fields

        return None

    def _render(self, nav_color: VtColors) -> None:
        """TODO"""

        restore_cursor()

        for idx, field in enumerate(self.all_fields):

            field_size = len(str(field.value))
            if self.tab_index == idx:
                MInputUtils.mi_print(self.max_label_length, field.label, MenuInput.SELECTED_BG)
                self.cur_field = field
            else:
                MInputUtils.mi_print(self.max_label_length, field.label)

            self._buffer_pos(field_size, idx)

            if field.itype == InputType.TEXT:
                MInputUtils.mi_print(self.max_value_length, field.value)
            elif field.itype == InputType.PASSWORD:
                MInputUtils.mi_print(self.max_value_length, '*' * field_size)
            elif field.itype == InputType.CHECKBOX:
                MInputUtils.mi_print(
                    self.max_value_length - 1, ' ', str(FormIcons.CHECK_SQUARE)
                    if field.value else str(FormIcons.UNCHECK_SQUARE))
            elif field.itype == InputType.SELECT:
                field_size = 1
                if field.value:
                    mat = re.search(r'.*\|?<(.+)>\|?.*', field.value)
                    sel_value = mat.group(1) if mat else field.value.split('|')[0]
                    MInputUtils.mi_print(self.max_value_length, f'{sel_value}')
            elif field.itype == InputType.MASKED:
                value, mask = MInputUtils.unpack_masked(str(field.value))
                MInputUtils.mi_print(self.max_value_length, MInputUtils.over_masked(value, mask))

            # Remaining/max characters
            self._render_details(field, field_size)

        sysout(f"\n{nav_color.placeholder()}{self.NAV_BAR}", end='')
        self.re_render = False

    def _buffer_pos(self, field_size: int, idx: int) -> None:
        """TODO"""

        # Buffering the all positions to avoid calling get_cursor_pos over and over
        f_pos = get_cursor_position() if self.all_pos[idx] == (0, 0) else self.all_pos[idx]
        if f_pos:
            self.all_pos[idx] = f_pos
            if self.tab_index == idx:
                self.cur_row = f_pos[0]
                self.cur_col = f_pos[1] + field_size

    def _render_details(self, field: FormField, field_size: int) -> None:
        """TODO"""

        # Print details about total/remaining field characters
        padding = 1 - len(str(self.max_detail_length / 2))
        fmt = '{:<3}{:>' + str(padding) + '}/{:<' + str(padding) + '} %MOD(0)%'
        if field.itype == InputType.SELECT:
            idx, _ = MInputUtils.get_selected(field.value)
            sysout(fmt.format(field.icon, idx + 1 if idx >= 0 else 1, len(field.value.split('|'))))
        elif field.itype == InputType.MASKED:
            value, _ = MInputUtils.unpack_masked(str(field.value))
            sysout(fmt.format(field.icon, len(value), field.max_length))
        else:
            sysout(fmt.format(field.icon, field_size, field.max_length))

    # pylint: disable=too-many-branches
    def _nav_input(self) -> chr:
        """TODO"""

        length = len(self.all_fields)
        keypress = Keyboard.read_keystroke()

        if keypress:
            if keypress == Keyboard.VK_ESC:
                self.done = True
            elif keypress in [Keyboard.VK_TAB, Keyboard.VK_DOWN]:  # Handle TAB and Cursor down
                self.tab_index = min(length - 1, self.tab_index + 1)
            elif keypress in [Keyboard.VK_SHIFT_TAB, Keyboard.VK_UP]:  # Handle Shift + TAB and Cursor up
                self.tab_index = max(0, self.tab_index - 1)
            elif keypress == Keyboard.VK_BACKSPACE:  # Handle backspace
                if not self.cur_field.can_write():
                    self._display_error('This field is read only !')
                else:
                    self._handle_backspace()
            elif keypress.isalnum() or keypress.ispunct() or keypress == Keyboard.VK_SPACE:  # Handle an input
                if not self.cur_field.can_write():
                    self._display_error('This field is read only !')
                else:
                    self._handle_input(keypress)
            elif keypress == Keyboard.VK_ENTER:  # Validate & Save form and exit
                invalid = next((field for field in self.all_fields if not field.validate(field.value)), None)
                if invalid:
                    keypress = None
                    idx = self.all_fields.index(invalid)
                    self.cur_row = self.all_pos[idx][0]
                    self.tab_index = idx
                    self._display_error("Form field is not valid: " + str(invalid))
                else:
                    for idx, field in enumerate(self.all_fields):
                        if field.itype == InputType.MASKED:
                            field.value = field.value.split('|')[0]
                        elif field.itype == InputType.CHECKBOX:
                            field.value = bool(field.value)
                        elif field.itype == InputType.SELECT:
                            _, field.value = MInputUtils.get_selected(field.value)
                    return keypress

        self.re_render = True
        return keypress

    def _handle_input(self, keypress: chr) -> None:
        """TODO"""

        if self.cur_field.itype == InputType.CHECKBOX:
            if keypress == Keyboard.VK_SPACE:
                self.cur_field.value = 1 if not self.cur_field.value else 0
        elif self.cur_field.itype == InputType.SELECT:
            if keypress == Keyboard.VK_SPACE:
                if self.cur_field.value:
                    self.cur_field.value = MInputUtils.toggle_selected(str(self.cur_field.value))
        elif self.cur_field.itype == InputType.MASKED:
            value, mask = MInputUtils.unpack_masked(str(self.cur_field.value))
            if len(value) < self.cur_field.max_length:
                try:
                    self.cur_field.value = MInputUtils.append_masked(value, mask, keypress.value)
                except InvalidInputError as err:
                    self._display_error(f"{str(err)}")
        else:
            if len(str(self.cur_field.value)) < self.cur_field.max_length:
                if self.cur_field.validate(keypress.value):
                    self.cur_field.value = str(self.cur_field.value) + str(keypress.value)
                else:
                    self._display_error(
                        f"This {self.cur_field.itype} field only accept {self.cur_field.validator} !")

    def _handle_backspace(self) -> None:
        """TODO"""

        if self.cur_field.itype == InputType.MASKED:
            value, mask = MInputUtils.unpack_masked(str(self.cur_field.value))
            value = value[:-1]
            while mask[len(value) - 1] not in ['#', '@', '*']:
                value = value[:-1]
            self.cur_field.value = f"{value}|{mask}"
        elif self.cur_field.itype not in [InputType.CHECKBOX, InputType.SELECT]:
            if self.cur_field.can_write() and len(str(self.cur_field.value)) >= 1:
                self.cur_field.value = str(self.cur_field.value)[:-1]
            elif not self.cur_field.can_write():
                self._display_error('This field is read only !')

    def _display_error(self, err_msg) -> None:
        """TODO"""

        set_enable_echo(False)
        err_pos = self.max_label_length + self.max_value_length + self.max_detail_length + 12
        vt_print(f"%CUP({self.cur_row};{err_pos})%")
        syserr(f"{FormIcons.ERROR_CIRCLE}  {err_msg}", end='')
        time.sleep(max(2, int(len(err_msg) / 25)))
        set_enable_echo()
        vt_print(f"%CUP({self.cur_row};{err_pos})%%EL0%")  # Remove the message after the timeout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: application.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import sys
import traceback
from typing import Callable

from hspylib.core.tools.commons import syserr


class ExitHooks:
    """TODO"""

    def __init__(self, cleanup: Callable = None):
        self._orig_exit = None
        self._exit_code = None
        self._exception = None
        self._traceback = None
        self._cleanup = cleanup

    def hook(self) -> None:
        """TODO"""
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exception_handler

    def exit(self, code=0) -> None:
        """TODO"""
        self._exit_code = code
        self._orig_exit(code)

    def exception_handler(self, exc_type: TypeError, exc: BaseException,
                          tb: traceback) -> None:  # pylint: disable=unused-argument
        """TODO"""
        self._exception = exc
        self._traceback = tb
        tb = "\t".join(traceback.format_exception(exc_type, exc, self._traceback))
        syserr(tb)
        if self._cleanup:
            self._cleanup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
   @file: zoned_datetime.py
  @created: Thu, 20 Sep 2022
   @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
     @site: https://github.com/yorevs/homesetup
  @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from datetime import datetime, timezone

from hspylib.core.enums.enumeration import Enumeration

# Date and time formats
# @formatter:off
ISO_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S%z'
DATETIME_FORMAT     = '%Y-%m-%d %H:%M:%S'
DATE_FORMAT         = '%Y-%m-%d'
TIME_FORMAT         = '%H:%M:%S'
# @formatter:on

class ZonedDatetime(Enumeration):

    # @formatter:off
    LOCAL   = 'localtime'
    UTC     = 'UTC'
    # @formatter:on

    def datetime(self) -> datetime:
        """ Return the current datetime"""
        now_dt = datetime.now(timezone.utc)
        if self == ZonedDatetime.LOCAL:
            now_dt = now_dt.astimezone()

        return now_dt

    def now(self, date_time_fmt: str = DATETIME_FORMAT) -> str:
        """ Return a formatted datetime string
        Ref. https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
        """

        return self.datetime().strftime(date_time_fmt)

    def now_ms(self) -> int:
        """ Return the datetime timestamp """

        return int(self.datetime().timestamp())


def now(date_time_fmt: str = DATETIME_FORMAT) -> str:
    return ZonedDatetime.LOCAL.now(date_time_fmt)


def now_ms() -> int:
    return ZonedDatetime.LOCAL.now_ms()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.qt
      @file: qt_application.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import sys
from pathlib import Path
from typing import TypeVar

from PyQt5.QtGui import QFont, QFontDatabase, QIcon
from PyQt5.QtWidgets import QApplication

from hspylib.core.preconditions import check_argument, check_state
from hspylib.core.tools.text_tools import titlecase
from hspylib.modules.cli.application.application import Application
from hspylib.modules.cli.application.version import Version

V = TypeVar('V', bound='QWidget')


class QtApplication(Application):

    def __init__(
        self,
        main_view: V,
        name: str,
        version: Version,
        description: str = None,
        usage: str = None,
        epilog: str = None,
        resource_dir: str = None,
        log_dir: str = None):
        super().__init__(name, version, description, usage, epilog, resource_dir, log_dir)
        app_title = titlecase(name)
        self.qapp = QApplication(sys.argv)
        self.main_view = main_view()
        self.main_view.window.setWindowTitle(f"{app_title} v{str(version)}")
        self.qapp.setApplicationDisplayName(app_title)
        self.qapp.setApplicationName(name)
        self.qapp.setApplicationVersion(str(version))
        self.qapp.setQuitOnLastWindowClosed(True)

    def _main(self, *params, **kwargs) -> int:
        """Execute the application's main statements"""
        self.main_view.show()
        a = self.qapp.exec_()
        return 0

    def _cleanup(self) -> None:
        QApplication.exit()

    def set_application_font(self, font_path: Path) -> None:
        """TODO"""
        check_argument(font_path.exists(), f'Could not find font at: {str(font_path)}')
        font_id = QFontDatabase.addApplicationFont(str(font_path))
        families = QFontDatabase.applicationFontFamilies(font_id)
        check_state(families is not None and len(families) == 1)
        self.qapp.setFont(QFont(families[0], 14))

    def set_application_icon(self, icon_path: Path) -> None:
        """TODO"""
        check_argument(icon_path.exists(), f'Could not find icon file at: {str(icon_path)}')
        self.qapp.setWindowIcon(QIcon(str(icon_path)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.main.hspylib.modules.qt.promotions
      @file: hframe.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QFrame

class HFrame(QFrame):
    """TODO"""

    keyPressed = pyqtSignal(int)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """TODO"""
        self.keyPressed.emit(event.key())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: dashboard_icons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.modules.cli.icons.font_awesome.awesome import Awesome

class DashboardIcons(Awesome):
    """
        Dashboard UI icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    # @formatter:off
    ADJUSTMENTS = '\uFB2D'      # שּׂ
    AGENDA = '\uF073'           # 
    ALERT = '\uF071'            # 
    ARCHIVE = '\uF01C'          # 
    BALLOON = '\uF075'          # 
    CAMERA = '\uF030'           # 
    CHAT = '\uF78B'             # 
    CLOUD_DOWNLOAD = '\uF0ED'   # 
    CLOUD_UPLOAD = '\uF0EE'     # 
    CODE = '\uFABF'             # 謹
    COMPASS = '\uF124'          # 
    DASHBOARD = '\uF00A'        # 
    DATABASE = '\uF1C0'         # 
    DELIVERED = '\uFB3D'        # ﬽
    DOCUMENT = '\uF016'         # 
    DOWNLOAD = '\uF019'         # 
    EMAIL = '\uF003'            # 
    ERROR = '\uF06A'            # 
    EXIT = '\uF08B'             # 
    EXPORT = '\uF5F6'           # 
    FAVORITE = '\uF02E'         # 
    FLAG = '\uF024'             # 
    FOLDER = '\uF74A'           # 
    FOLDER_OPEN = '\uF07C'      # 
    FORM = '\uF767'             # 
    FORMAT = '\uF9C5'           # 暈
    HISTORY = '\uF017'          # 
    HOME = '\uF015'             # 
    IDENTITY = '\uF2BB'         # 
    IMPORT = '\uF5F9'           # 
    KILL = '\uFB8A'             # ﮊ
    LIGHT_ON = '\uF834'         # 
    LIGHT_OFF = '\uF835'        # 
    LIKE = '\uF004'             # 
    LIST = '\uF03A'             # 
    LOCATION = '\uF450'         # 
    MOBILE = '\uF61F'           # 
    MOVIE = '\uF008'            # 
    MUSIC = '\uF3B5'            # 
    MUTE = '\uF86C'             # 
    NOTIFICATION = '\uF0F3'     # 
    PICTURE = '\uF03E'          # 
    PLAY = '\uF01D'             # 
    PLUGIN = '\uF930'           # 擄
    PLUG_IN = '\uF1E6'          # 
    PLUG_OUT = '\uF492'         # 
    PLUGGED = '\uFBA3'          # ﮣ
    POWER = '\uF011'            # 
    PRINTER = '\uF02F'          # 
    READ = '\uFB3C'             # לּ
    SAVE = '\uF0C7'             # 
    SEARCH = '\uF002'           # 
    SECRET = '\uF49C'           # 
    SEND = '\uF989'             # 黎
    SENT = '\uF1D8'             # 
    SETTINGS = '\uF013'         # 
    SHOPPING_CART = '\uF60F'    # 
    UN_MUTE = '\uF86B'          # 
    UNPLUGGED = '\uFBA4'        # ﮤ
    USER = '\uF007'             # 
    VERIFIED = '\uFC8F'         # ﲏ

    # @formatter:on


if __name__ == '__main__':
    DashboardIcons.demo_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.metaclass
      @file: singleton.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import sys
from typing import Type, TypeVar, Union

from hspylib.core.exception.exceptions import HSBaseException
from hspylib.core.preconditions import check_not_none

SINGLETON = TypeVar('SINGLETON', bound=Union[Type, 'Singleton'])


# pylint: disable=bad-mcs-classmethod-argument
class Singleton(Type):
    """Singleton pattern is a software design pattern that restricts the instantiation of a class to a singular
    instance. This metaclass enables a class to be singleton."""

    _instances = {}

    def __call__(self, *args, **kwargs):
        """Invoke the class constructor or return the instance if it exists."""
        if not Singleton.has_instance(self):
            try:
                instance = super(Singleton, self).__call__(*args, **kwargs)
                check_not_none(instance, f'Unable to create Singleton instance: {self}')
                setattr(self, 'INSTANCE',  instance)
                Singleton._instances[self.__name__] = instance
                log.debug('Created a new Singleton instance: %s.%s', self.__module__, self.__name__)
            except Exception as err:  # pylint: disable=broad-except
                raise HSBaseException(f"Failed to create singleton instance: '{self.__name__}'", err) from err
        return Singleton._instances[self.__name__]

    @classmethod
    def has_instance(cls, clazz: SINGLETON) -> bool:
        """Whether the class has an instance or not. """
        return clazz.__name__ in cls._instances

    @classmethod
    def del_instance(cls, clazz: SINGLETON) -> None:
        """Deletes the singleton instance. This method should be used only for testing purposes."""
        if any(m in sys.modules.keys() for m in ['unittest', 'pytest']):
            if Singleton.has_instance(clazz):
                log.warning('Deleted an existing Singleton instance: %s.%s', cls.__module__, cls.__name__)
                del cls._instances[clazz.__name__]
                delattr(clazz, 'INSTANCE')
                del clazz
            else:
                raise TypeError(f"Instance not found: '{clazz.__name__}'")
        else:
            raise TypeError("This method is only available for testing purposes (cleanup).")



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.121
#
# Package: main.hspylib.addons.widman
"""Package initialization."""

__all__ = [
  'widget', 
  'widget_entry', 
  'widgets', 
  'widman'
]
__version__ = '0.11.121'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.enum
      @file: http_code.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from http import HTTPStatus as httpCode

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.tools.text_tools import titlecase


class HttpCode(Enumeration):
    """HTTP status code constants defined by the proposed HTTP/1.1 specification at http://www.ietf.org/rfc/rfc2068.txt.
    Ref:http://www.cs.ait.ac.th/~on/O/oreilly/java-ent/servlet/appc_01.htm
    """

    # @formatter:off

    # 1xx Informational Codes
    CONTINUE                            = httpCode.CONTINUE
    SWITCHING_PROTOCOLS                 = httpCode.SWITCHING_PROTOCOLS
    PROCESSING                          = httpCode.PROCESSING
    EARLY_HINTS                         = httpCode.EARLY_HINTS

    # 2xx Success Codes
    OK                                  = httpCode.OK
    CREATED                             = httpCode.CREATED
    ACCEPTED                            = httpCode.ACCEPTED
    NON_AUTHORITATIVE_INFORMATION       = httpCode.NON_AUTHORITATIVE_INFORMATION
    NO_CONTENT                          = httpCode.NO_CONTENT
    RESET_CONTENT                       = httpCode.RESET_CONTENT
    PARTIAL_CONTENT                     = httpCode.PARTIAL_CONTENT
    MULTI_STATUS                        = httpCode.MULTI_STATUS
    ALREADY_REPORTED                    = httpCode.ALREADY_REPORTED
    IM_USED                             = httpCode.IM_USED

    # 3xx Redirection Codes
    MULTIPLE_CHOICES                    = httpCode.MULTIPLE_CHOICES
    MOVED_PERMANENTLY                   = httpCode.MOVED_PERMANENTLY
    FOUND                               = httpCode.FOUND
    SEE_OTHER                           = httpCode.SEE_OTHER
    NOT_MODIFIED                        = httpCode.NOT_MODIFIED
    USE_PROXY                           = httpCode.USE_PROXY
    TEMPORARY_REDIRECT                  = httpCode.TEMPORARY_REDIRECT
    PERMANENT_REDIRECT                  = httpCode.PERMANENT_REDIRECT

    # 4xx Client Error Codes
    BAD_REQUEST                         = httpCode.BAD_REQUEST
    UNAUTHORIZED                        = httpCode.UNAUTHORIZED
    PAYMENT_REQUIRED                    = httpCode.PAYMENT_REQUIRED
    FORBIDDEN                           = httpCode.FORBIDDEN
    NOT_FOUND                           = httpCode.NOT_FOUND
    METHOD_NOT_ALLOWED                  = httpCode.METHOD_NOT_ALLOWED
    NOT_ACCEPTABLE                      = httpCode.NOT_ACCEPTABLE
    PROXY_AUTHENTICATION_REQUIRED       = httpCode.PROXY_AUTHENTICATION_REQUIRED
    REQUEST_TIMEOUT                     = httpCode.REQUEST_TIMEOUT
    CONFLICT                            = httpCode.CONFLICT
    GONE                                = httpCode.GONE
    LENGTH_REQUIRED                     = httpCode.LENGTH_REQUIRED
    PRECONDITION_FAILED                 = httpCode.PRECONDITION_FAILED
    PAYLOAD_TOO_LARGE                   = 413
    REQUEST_URI_TOO_LONG                = httpCode.REQUEST_URI_TOO_LONG
    UNSUPPORTED_MEDIA_TYPE              = httpCode.UNSUPPORTED_MEDIA_TYPE
    REQUESTED_RANGE_NOT_SATISFIABLE     = httpCode.REQUESTED_RANGE_NOT_SATISFIABLE
    EXPECTATION_FAILED                  = httpCode.EXPECTATION_FAILED
    UNPROCESSABLE_ENTITY                = httpCode.UNPROCESSABLE_ENTITY
    LOCKED                              = httpCode.LOCKED
    FAILED_DEPENDENCY                   = httpCode.FAILED_DEPENDENCY
    UPGRADE_REQUIRED                    = httpCode.UPGRADE_REQUIRED
    PRECONDITION_REQUIRED               = httpCode.PRECONDITION_REQUIRED
    TOO_MANY_REQUESTS                   = httpCode.TOO_MANY_REQUESTS
    REQUEST_HEADER_FIELDS_TOO_LARGE     = httpCode.REQUEST_HEADER_FIELDS_TOO_LARGE
    CONNECTION_CLOSED_WITHOUT_RESPONSE  = 444
    UNAVAILABLE_FOR_LEGAL_REASONS       = 451
    CLIENT_CLOSED_REQUEST               = 499

    # 5xx Server Error
    INTERNAL_SERVER_ERROR               = httpCode.INTERNAL_SERVER_ERROR
    NOT_IMPLEMENTED                     = httpCode.NOT_IMPLEMENTED
    BAD_GATEWAY                         = httpCode.BAD_GATEWAY
    SERVICE_UNAVAILABLE                 = httpCode.SERVICE_UNAVAILABLE
    GATEWAY_TIMEOUT                     = httpCode.GATEWAY_TIMEOUT
    HTTP_VERSION_NOT_SUPPORTED          = httpCode.HTTP_VERSION_NOT_SUPPORTED
    VARIANT_ALSO_NEGOTIATES             = httpCode.VARIANT_ALSO_NEGOTIATES
    INSUFFICIENT_STORAGE                = httpCode.INSUFFICIENT_STORAGE
    LOOP_DETECTED                       = httpCode.LOOP_DETECTED
    NOT_EXTENDED                        = httpCode.NOT_EXTENDED
    NETWORK_AUTHENTICATION_REQUIRED     = httpCode.NETWORK_AUTHENTICATION_REQUIRED
    NETWORK_CONNECT_TIMEOUT_ERROR       = 599

    # @formatter:on

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"({self.value}) {titlecase(self.name)}"




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.application
      @file: application.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import argparse
import atexit
import logging as log
import os
import signal
import sys
import traceback
from textwrap import dedent
from typing import Optional, Union

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.exception.exceptions import InvalidArgumentError, InvalidOptionError, \
    InvalidStateError
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_state
from hspylib.core.tools.commons import log_init, sysout
from hspylib.core.tools.text_tools import camelcase
from hspylib.core.tools.zoned_datetime import now
from hspylib.modules.cli.application.argparse.argument_parser import HSArgumentParser
from hspylib.modules.cli.application.argparse.arguments_builder import ArgumentsBuilder
from hspylib.modules.cli.application.argparse.chained_arguments_builder import ChainedArgumentsBuilder
from hspylib.modules.cli.application.argparse.options_builder import OptionsBuilder
from hspylib.modules.cli.application.exit_hooks import ExitHooks
from hspylib.modules.cli.application.version import Version


class Application(metaclass=Singleton):
    """HSPyLib application framework"""

    INSTANCE = None

    @staticmethod
    def exit(signum=0, frame=None, clear_screen: bool = False) -> None:
        """
        Handle interruptions to shutdown gracefully
        :param signum: The signal number or the exit http_code
        :param frame: The frame raised by the signal
        :param clear_screen: Whether to clean the screen before execution or not
        """
        if frame is not None:
            log.warning('Signal handler hooked signum=%d frame=%s', signum, frame)
            exit_code = 3
        else:
            log.info('Exit handler called')
            exit_code = signum
        if clear_screen:
            sysout('%ED2%%HOM%')
        sys.exit(exit_code)

    def __init__(
        self,
        name: str,
        version: Version,
        description: str = None,
        usage: str = None,
        epilog: str = None,
        resource_dir: str = None,
        log_dir: str = None):

        signal.signal(signal.SIGINT, self.exit)
        signal.signal(signal.SIGTERM, self.exit)

        self.exit_hooks = ExitHooks(self._cleanup)
        self.exit_hooks.hook()
        self._run_dir = os.getcwd()
        self._app_name = name
        self._app_version = version
        self._app_description = description
        self._arg_parser = HSArgumentParser(
            exit_on_error=False, prog=name, allow_abbrev=False,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=dedent(description or ''), usage=usage,
            epilog=dedent(epilog or ''))
        self._arg_parser.add_argument(
            '-v', '--version', action='version', version=f"%(prog)s v{self._app_version}")
        self._args = {}
        self._exit_code = 0

        # Initialize application configs
        if os.path.exists(f'{resource_dir}'):
            self.configs = AppConfigs(resource_dir=resource_dir)
        elif not resource_dir and os.path.exists(f'{self._run_dir}/resources/application.properties'):
            self.configs = AppConfigs(resource_dir=f'{self._run_dir}/resources')
        else:
            log.debug(f'Resource dir \"{resource_dir or "<none>"}\" not found. AppConfigs will not be available!')

        # Initialize application logs
        self._log_file = f"{log_dir or os.getenv('LOG_DIR', os.getcwd())}/{camelcase(name)}.log"
        check_state(log_init(self._log_file), "Unable to initialize logging. log_file={}", self._log_file)

    def run(self, *params, **kwargs) -> None:
        """Main entry point handler"""
        today = now()
        log.info('Application %s started %s', self._app_name, today)
        try:
            atexit.register(self._cleanup)
            self._setup_arguments()
            self._args = self._arg_parser.parse_args(*params)
            log.debug('Command line arguments: %s', str(self._args))
            self._exit_code = self._main(*params, **kwargs)
        except (InvalidOptionError, InvalidArgumentError) as err:
            self.usage(exit_code=1, no_exit=True)
            log.error('Run failed %s => %s', today, err)
            raise err  # Re-Raise the exception so upper level layers can catch
        except InvalidStateError as err:
            log.error('Execution failed %s => %s', today, err)
            raise err  # Re-Raise the exception so upper level layers can catch
        except Exception as err:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exc(file=sys.stderr)
            self._exit_code = exc_value
            raise err  # Re-Raise the exception so upper level layers can catch
        finally:
            log.info('Application %s finished %s', self._app_name, today)
            if 'no_exit' not in kwargs:
                self.exit(self._exit_code)

    def usage(self, exit_code: int = 0, no_exit: bool = False) -> None:
        """Display the usage message and exit with the specified http_code ( or zero as default )
        :param no_exit: Do no exit the application on usage call
        :param exit_code: The exit http_code
        """
        self._arg_parser.print_help(sys.stderr if exit_code != 0 else sys.stdout)
        if not no_exit:
            self.exit(exit_code)

    def version(self) -> str:
        """Return the application version"""
        return str(self._app_version)

    def name(self) -> str:
        """Return the application name"""
        return self._app_name

    def get_arg(self, arg_name: str) -> Optional[Union[str, list]]:
        """Get the argument value named by arg_name"""
        return getattr(self._args, arg_name) \
            if self._args and hasattr(self._args, arg_name) \
            else None

    def _setup_arguments(self) -> None:
        """Initialize application parameters and options"""

    def _with_options(self) -> 'OptionsBuilder':
        """TODO"""
        return OptionsBuilder(self._arg_parser)

    def _with_arguments(self) -> 'ArgumentsBuilder':
        """TODO"""
        return ArgumentsBuilder(self._arg_parser)

    def _with_chained_args(self, subcommand_name: str, subcommand_help: str = None) -> 'ChainedArgumentsBuilder':
        return ChainedArgumentsBuilder(self._arg_parser, subcommand_name, subcommand_help)

    def _main(self, *params, **kwargs) -> int:
        """Execute the application's main statements"""

    def _cleanup(self) -> None:
        """Execute http_code cleanup before exiting"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.enum
      @file: charset.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class Charset(Enumeration):
    """Enumeration to wrap all charset/encode types."""

    # @formatter:off

    ASCII           = 'ASCII'
    BIG5            = 'BIG5'
    BIG5HKSCS       = 'BIG5HKSCS'

    CP037           = 'CP037'
    CP424           = 'CP424'
    CP437           = 'CP437'
    CP500           = 'CP500'
    CP737           = 'CP737'
    CP775           = 'CP775'
    CP850           = 'CP850'
    CP852           = 'CP852'
    CP855           = 'CP855'
    CP856           = 'CP856'
    CP857           = 'CP857'
    CP860           = 'CP860'
    CP861           = 'CP861'
    CP862           = 'CP862'
    CP863           = 'CP863'
    CP864           = 'CP864'
    CP865           = 'CP865'
    CP866           = 'CP866'
    CP869           = 'CP869'
    CP874           = 'CP874'
    CP875           = 'CP875'
    CP932           = 'CP932'
    CP949           = 'CP949'
    CP950           = 'CP950'
    CP1006          = 'CP1006'
    CP1026          = 'CP1026'
    CP1140          = 'CP1140'
    CP1250          = 'CP1250'
    CP1251          = 'CP1251'
    CP1252          = 'CP1252'
    CP1253          = 'CP1253'
    CP1254          = 'CP1254'
    CP1255          = 'CP1255'
    CP1256          = 'CP1256'
    CP1257          = 'CP1257'
    CP1258          = 'CP1258'

    EUC_JP          = 'EUC-JP'
    EUC_JIS_2004    = 'EUC-JIS_2004'
    EUC_JISX0213    = 'EUC-JISX0213'
    EUC_KR          = 'EUC-KR'
    GB2312          = 'GB2312'
    GBK             = 'GBK'
    GB18030         = 'GB18030'
    HZ              = 'HZ'

    ISO2022_JP      = 'ISO-2022-JP'
    ISO2022_JP_1    = 'ISO-2022-JP-1'
    ISO2022_JP_2    = 'ISO-2022-JP-2'
    ISO2022_JP_2004 = 'ISO-2022-JP-2004'
    ISO2022_JP_3    = 'ISO-2022-JP-3'
    ISO2022_JP_EXT  = 'ISO-2022-JP_EXT'
    ISO2022_KR      = 'ISO-2022-KR'

    LATIN_1         = 'LATIN-1'
    ISO8859_1       = 'ISO-8859-1'
    ISO8859_2       = 'ISO-8859-2'
    ISO8859_3       = 'ISO-8859-3'
    ISO8859_4       = 'ISO-8859-4'
    ISO8859_5       = 'ISO-8859-5'
    ISO8859_6       = 'ISO-8859-6'
    ISO8859_7       = 'ISO-8859-7'
    ISO8859_8       = 'ISO-8859-8'
    ISO8859_9       = 'ISO-8859-9'
    ISO8859_10      = 'ISO-8859-10'
    ISO8859_13      = 'ISO-8859-13'
    ISO8859_14      = 'ISO-8859-14'
    ISO8859_15      = 'ISO-8859-15'

    JOHAB           = 'JOHAB'
    KOI8_R          = 'KOI8-R'
    KOI8_U          = 'KOI8-U'

    MAC_CYRILLIC    = 'MAC-CYRILLIC'
    MAC_GREEK       = 'MAC-GREEK'
    MAC_ICELAND     = 'MAC-ICELAND'
    MAC_LATIN2      = 'MAC-LATIN2'
    MAC_ROMAN       = 'MAC-ROMAN'
    MAC_TURKISH     = 'MAC-TURKISH'

    PTCP154         = 'PTCP154'
    SHIFT_JIS       = 'SHIFT-JIS'
    SHIFT_JIS_2004  = 'SHIFT-JIS_2004'
    SHIFT_JISX0213  = 'SHIFT-JISX0213'

    UTF_16          = 'UTF-16'
    UTF_16_BE       = 'UTF-16-BE'
    UTF_16_LE       = 'UTF-16-LE'
    UTF_7           = 'UTF-7'
    UTF_8           = 'UTF-8'

    # @formatter:on

    @property
    def value(self) -> str:
        return str(super().value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: app_icons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.modules.cli.icons.font_awesome.awesome import Awesome

class AppIcons(Awesome):
    """
        Application icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    # @formatter:off
    APPLE = '\uF534'            # 
    AWS = '\uF375'              # 
    CPLUSPLUS = '\uFB70'        # ﭱ
    DOCKER = '\uF308'           # 
    DROPBOX = '\uF6E2'          # 
    GIT = '\uF813'              # 
    GITHUB = '\uF408'           # 
    GOOGLE_DRIVE = '\uF3AA'     # 
    FACEBOOK = '\uF082'         # 
    FIREBASE = '\uF1D0'         # 
    IE = '\uF7FF'               # 
    JAVA = '\uF4E4'             # 
    LINKED_IN = '\uF08C'        # 
    PYTHON = '\uF81F'           # 
    STACK_OVERFLOW = '\uF9CB'   # 溜
    TWITTER = '\uF081'          # 
    VS_CODE = '\uFB0F'          # ﬏
    WHATSAPP = '\uFAA2'         # 甆
    # @formatter:on


if __name__ == '__main__':
    AppIcons.demo_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.icons.font_awesome
      @file: app_icons.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.modules.cli.icons.font_awesome.awesome import Awesome

class NavIcons(Awesome):
    """
        Navigation icons.
        Codes can be found here:
        - https://fontawesome.com/cheatsheet?from=io
    """

    # @formatter:off
    LEFT = '\u2190'         # ←
    UP = '\u2191'           # ↑
    RIGHT = '\u2192'        # →
    DOWN = '\u2193'         # ↓
    ENTER = '\u21b2'        # ↲
    TAB = '\u21b9'          # ↹
    # @formatter:on


if __name__ == '__main__':
    NavIcons.demo_icons()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.tools
      @file: namespace.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Any, Dict, Iterator, List, Optional, Tuple

from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.dict_tools import merge_iterables


class Namespace:

    @staticmethod
    def of(type_name: str, attributes: Dict[str, Any] | Tuple[Dict[str, Any]] | List[Dict[str, Any]]) -> 'Namespace':
        check_not_none(attributes)
        self = Namespace(type_name)
        self += attributes if isinstance(attributes, dict) else merge_iterables(attributes)
        return self

    def __init__(self, type_name: str = '', **kwargs) -> None:
        self.__name__ = type_name
        self._index = 0
        list(map(self.setattr, kwargs.keys(), kwargs.values()))

    def __key(self) -> Tuple[Any]:
        return tuple(self.values)

    def __str__(self) -> str:
        return f"{self.__name__}({', '.join([f'{a}={av}' for a, av in zip(self.attributes, self.values)])})"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other: 'Namespace') -> bool:
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return NotImplemented

    def __getitem__(self, attribute_name: str) -> Any:
        return getattr(self, attribute_name)

    def __len__(self):
        return len(self.attributes)

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        self._index = 0
        return self

    def __next__(self) -> Tuple[str, Any]:
        if self._index < len(self):
            item = self.at(self._index)
            self._index += 1
            return item
        raise StopIteration

    def __contains__(self, attribute: str):
        return hasattr(self, attribute)

    def __iadd__(self, attribute: Dict[str, Any] | 'Namespace') -> 'Namespace':
        for a, av in attribute.items():
            self.setattr(a, av)
        return self

    def __add__(self, other: Dict[str, Any] | 'Namespace') -> 'Namespace':
        self.__name__ += '.' + other.__name__
        return self.__iadd__(other)

    def setattr(self, name: str, value: Any) -> 'Namespace':
        if name.startswith(('_', '__')):
            raise NameError("Attribute names can't start with '_' or '__'")
        setattr(self, name, value)
        return self

    def at(self, index: int) -> Optional[Tuple[str, Any]]:
        return (self.attributes[index], self.values[index]) if index < len(self) else None

    def items(self) -> Iterator[Tuple[str, Any]]:
        return iter(self)

    @property
    def attributes(self) -> Tuple[str]:
        attrs = list(filter(lambda name: self[name] is not None and not name.startswith(('_', '__')), vars(self)))
        return tuple([a for a in attrs])

    @property
    def values(self) -> Tuple[Any]:
        return tuple(filter(lambda v: v is not None, [getattr(self, a) for a in self.attributes]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db
      @file: db_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import contextlib
from abc import abstractmethod
from typing import Any, Generic, Iterable, Optional, Tuple, TypeVar

from retry import retry

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.crud_repository import CrudRepository
from hspylib.core.datasource.db_configuration import DBConfiguration

Connection = TypeVar('Connection', bound=Any)
Cursor = TypeVar('Cursor', bound=Any)
Session = TypeVar('Session', bound=Any)
ResultSet = TypeVar('ResultSet', bound=Iterable)
E = TypeVar('E', bound=CrudEntity)
C = TypeVar('C', bound=DBConfiguration)


class DBRepository(Generic[E, C], CrudRepository[E]):
    """TODO"""

    def __init__(self, config: C):
        super().__init__()
        self._config = config

    def __str__(self):
        return f"{self.logname}/{self.table_name()} -> {self.info}"

    def __repr__(self):
        return str(self)

    @property
    def config(self) -> C:
        return self._config

    @property
    def info(self) -> str:
        return f"{self.username or ''}@{self.hostname or ''}:{self.port or 0}/{self.database or ''}"

    @property
    def hostname(self) -> str:
        return self._config.hostname

    @property
    def port(self) -> int:
        return self._config.port

    @property
    def username(self) -> str:
        return self._config.username

    @property
    def password(self) -> str:
        return self._config.password

    @property
    def database(self) -> str:
        return self._config.database

    @abstractmethod
    @retry(tries=3, delay=2, backoff=3, max_delay=30)
    def _create_session(self) -> Tuple[Connection, Cursor]:
        """TODO"""

    @abstractmethod
    @contextlib.contextmanager
    def _session(self) -> Session:
        """Create a database session."""

    @abstractmethod
    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        """Execute a SQL statement."""

    @abstractmethod
    def table_name(self) -> str:
        """TODO"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui
      @file: menu_utils.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import time
from abc import ABC
from typing import Any, Callable, Optional

from hspylib.core.exception.exceptions import InputAbortedError
from hspylib.core.tools.commons import syserr, sysout
from hspylib.core.tools.validator import Validator
from hspylib.modules.cli.vt100.vt_colors import VtColors


class MenuUtils(ABC):
    """TODO"""

    @staticmethod
    def print_error(
        message: str,
        argument: Any = None,
        wait_interval: int = 2) -> None:
        """TODO"""

        syserr(f"### Error: {message} \"{argument or ''}\"")
        time.sleep(wait_interval)
        sysout('%CUU(2)%%ED0%', end='')

    @staticmethod
    def print_warning(
        message: str,
        argument: str = None,
        wait_interval: int = 2,
        color: VtColors = VtColors.YELLOW) -> None:
        """TODO"""

        sysout(f"{color.placeholder()}### Warn: {message} \"{argument or ''}\"")
        time.sleep(wait_interval)
        sysout('%CUU(2)%%ED0%', end='')

    @staticmethod
    def prompt(
        prompt_msg: str = '',
        validator: Callable = None,
        default_value: Any = None,
        any_key: bool = False,
        on_blank_abort: bool = True,
        color: VtColors = VtColors.GREEN,
        end: str = ': ') -> Optional[Any]:
        """TODO"""

        valid = False
        input_data = None

        while not valid:
            try:
                colorized = VtColors.colorize(
                    f"{color.placeholder()}{prompt_msg}{f'[{default_value}]' if default_value else ''}{end}{VtColors.NC.placeholder()}"
                )
                input_data = input(colorized)
                if not validator:
                    valid = True
                elif default_value:
                    return default_value
                elif any_key:
                    return None
                elif Validator.is_not_blank(input_data):
                    valid, msg = validator(input_data)
                    if not valid:
                        MenuUtils.print_error(f"{msg}: ", input_data)
                else:
                    if not on_blank_abort:
                        MenuUtils.print_error("Input can't be empty: ", input_data)
                    else:
                        raise InputAbortedError('Input process was aborted')
            except EOFError as err:
                MenuUtils.print_error("Input failed: ", str(err))
                break

        return input_data

    @staticmethod
    def wait_enter(
        wait_msg: str = 'Press [Enter] to continue ...',
        color: VtColors = VtColors.YELLOW) -> None:
        """TODO"""
        MenuUtils.prompt(wait_msg, any_key=True, color=color, end='')

    @staticmethod
    def title(title_str: str, color: VtColors = VtColors.YELLOW) -> None:
        """TODO"""
        sysout(f"%ED2%%HOM%\n{color.placeholder()}{title_str}\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: hspylib
   @package: hspylib.main.hspylib
      @file: __classpath__.py
   @created: Wed, 8 Jun 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.metaclass.classpath import Classpath
from hspylib.core.tools.commons import get_path, run_dir


class _Classpath(Classpath):
    """TODO"""

    def __init__(self) -> None:
        super().__init__(
            get_path(__file__),
            get_path(run_dir()),
            (get_path(__file__) / "resources"))


# Instantiate the classpath singleton
_Classpath()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.model
      @file: crud_entity.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import json
from typing import Iterable, Optional
from uuid import UUID

from hspylib.core.datasource.identity import Identity
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.namespace import Namespace


class CrudEntity(Namespace):
    """Generic entity type."""

    def __init__(self, entity_id: Optional[Identity] = None, **kwargs):
        super().__init__(self.__class__.__name__, **kwargs)
        check_argument(entity_id is None or isinstance(entity_id, Identity))
        self._identity = entity_id if entity_id else Identity.auto()
        self.__iadd__(self._identity.as_dict())

    @property
    def identity(self) -> Identity:
        return self._identity

    def as_dict(self) -> dict:
        """TODO"""

        ret_dict = {}
        for key, value in zip(self.attributes, self.values):
            if isinstance(value, bool):
                ret_dict[key] = bool(value)
            elif isinstance(value, int):
                ret_dict[key] = int(value)
            elif isinstance(value, float):
                ret_dict[key] = float(value)
            elif isinstance(value, (str, UUID)):
                ret_dict[key] = str(value)
            else:
                ret_dict[key] = value.__dict__ if value else {}
        return ret_dict

    def as_json(self, indent: int = None) -> str:
        """TODO"""
        return json.dumps(self.as_dict(), indent=indent)

    def as_column_set(
        self,
        separator: str = ', ',
        prefix: Optional[str] = None,
        exclude: Optional[Iterable[str]] = None) -> str:
        """TODO"""

        column_set = []
        exclude = exclude or []
        list(map(
            lambda key: column_set.append(f"{key} = {(prefix or '') + key}"),
            filter(lambda a: a not in exclude, self.attributes)
        ))

        return separator.join(column_set)

    def as_columns(self, separator: str = ', ') -> str:
        """TODO"""

        columns = []
        for key in self.attributes:
            columns.append(f"{key}".replace("'", ""))

        return separator.join(columns)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db.firebase
      @file: firebase_configuration.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import os
from os.path import basename
from textwrap import dedent
from typing import Optional

from requests.structures import CaseInsensitiveDict

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.enums.charset import Charset
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.preconditions import check_argument, check_not_none, check_state
from hspylib.core.tools.commons import dirname
from hspylib.modules.fetch.fetch import get


class FirebaseConfiguration(AppConfigs):
    """Represents a Firebase project configuration"""

    CONFIG_FORMAT = dedent("""
    # Your Firebase configuration:
    # --------------------------
    UID={uid}
    PROJECT_ID={project_id}
    EMAIL={email}
    DATABASE={database}
    BASE_URL={base_url}
    """)

    REQUIRED_SETTINGS = ['UID', 'PROJECT_ID', 'EMAIL', 'DATABASE']

    @staticmethod
    def of(
        resource_dir: str,
        filename: str,
        config_dict: CaseInsensitiveDict) -> 'FirebaseConfiguration':
        """Create a Firebase config from a case insensitive dictionary."""

        check_state(
            all(setting in config_dict for setting in FirebaseConfiguration.REQUIRED_SETTINGS),
            f"Invalid configuration file.\nMust contain required settings: {FirebaseConfiguration.REQUIRED_SETTINGS}")
        return FirebaseConfiguration(
            resource_dir,
            filename,
            config_dict['UID'],
            config_dict['PROJECT_ID'],
            config_dict['EMAIL'],
            config_dict['DATABASE']
        )

    @staticmethod
    def of_file(filename: str, encoding: str = str(Charset.UTF_8)) -> 'FirebaseConfiguration':
        """Create a Firebase config from a config file."""
        check_argument(os.path.exists(filename), f"Config file does not exist: {filename}")

        with open(filename, encoding=encoding or str(Charset.UTF_8)) as f_config:
            cfg = CaseInsensitiveDict()
            for line in f_config:
                line = line.strip()
                if line.startswith("#") or "=" not in line:
                    continue
                key, value = line.split("=", 1)
                if key in FirebaseConfiguration.REQUIRED_SETTINGS:
                    cfg[key.strip()] = value.strip()

            return FirebaseConfiguration.of(dirname(filename), basename(filename), cfg)

    def __init__(
        self,
        resource_dir: str,
        filename: str,
        uid: str,
        project_id: str,
        email: str,
        database: Optional[str] = None,
        profile: Optional[str] = None):

        super().__init__(resource_dir, filename, profile)
        self._valid = None
        self._base_url = self['datasource.base.url']
        self._scheme = self['datasource.scheme'] or 'https'
        self._hostname = self['datasource.hostname'] or 'firebaseio.com'
        self._port = self['datasource.port'] or 443 if self._scheme.endswith('s') else 80
        self._uid = uid or self['firebase.user.uid']
        self._project_id = project_id or self['firebase.project.id']
        self._email = email or self['firebase.email']
        self._database = database or self['firebase.database']

    def __str__(self) -> str:
        return '' if not hasattr(self, '_uid') else self.CONFIG_FORMAT.format(
            uid=self._uid or 'not-set',
            project_id=self._project_id or 'not-set',
            email=self._email or 'not-set',
            database=self.database or 'not-set',
            base_url=self.base_url or 'not-set'
        )

    @property
    def hostname(self) -> str:
        return f"{self.project_id}.{self._hostname}"

    @property
    def scheme(self) -> str:
        return self._scheme

    @property
    def port(self) -> int:
        return self._port

    @property
    def uid(self) -> str:
        return self._uid

    @property
    def project_id(self) -> str:
        return self._project_id

    @property
    def email(self) -> str:
        return self._email

    @property
    def database(self) -> Optional[str]:
        return self._database

    @property
    def base_url(self) -> str:
        return self._base_url or f'{self.scheme}://{self.hostname}:{self.port}/{self.database}'

    def validate_config(self) -> bool:
        """Validate the current configuration"""
        check_not_none(self._uid, "User UID must be defined")
        check_not_none(self._project_id, "Project ID must be defined")
        check_not_none(self._email, "Email must be defined")
        response = get(f'{self.base_url}.json')
        self._valid = response is not None and response.status_code == HttpCode.OK

        return self._valid

    def url(self, db_alias: Optional[str] = '') -> str:
        """Return the url for an element at the Firebase webapp."""
        final_alias_url = db_alias.replace(' ', '%20').replace('.', '/')
        return f'{self.base_url}/{final_alias_url}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.exception
      @file: exceptions.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import logging as log
import os
import sys


class HSBaseException(Exception):
    """This is a generic exception and should not be raised. It may be inherited instead"""

    def __init__(self, message: str, cause: Exception = None):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        if exc_type and exc_obj and exc_tb:
            filename = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            err_cause = str(cause) if cause else ''
            fmt_msg = f"### {message} :{err_cause}: (File {filename}, Line {exc_tb.tb_lineno})"
        else:
            fmt_msg = message
        super().__init__(fmt_msg)
        log.error(fmt_msg)


class ApplicationError(HSBaseException):
    """Raised when an entity is not found"""


class EntityNotFoundError(HSBaseException):
    """Raised when an entity is not found"""


class InputAbortedError(HSBaseException):
    """Raised when an input method is aborted"""


class ProgrammingError(HSBaseException):
    """Exception raised for programming errors, e.g. table not found
    or already exists, syntax error in the SQL statement, wrong number
    of parameters specified, etc."""


class InvalidOptionError(HSBaseException):
    """Raised when an invalid option is received by the application"""


class DatabaseConnectionError(HSBaseException):
    """Raised when all connection attempts to the database exhausted"""


class DatabaseError(HSBaseException):
    """Raised when database failed to execute/create-session"""


class WidgetNotFoundError(HSBaseException):
    """Raised when widget is not found on any of the widget paths"""


class WidgetExecutionError(HSBaseException):
    """Raised when widget failed to execute"""


class KeyboardInputError(HSBaseException):
    """Raised when an invalid keystroke input is provided"""


class ResourceNotFoundError(HSBaseException):
    """Raised when resource path is not found"""


class SourceNotFoundError(HSBaseException):
    """Raised when source path is not found"""


class InvalidInputError(Exception):
    """Raised when an invalid input is provided """


class InvalidArgumentError(Exception):
    """Raised when an invalid argument is received by the application"""


class InvalidStateError(Exception):
    """Raised when an invalid state is detected by the application"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file

   @project: HSPyLib
   @package: main.addons.widman.widgets
      @file: widget_punch.py
   @created: Thu, 20 Sep 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import argparse
import os
import re
import sys
from textwrap import dedent
from typing import List

from hspylib.addons.widman.widget import Widget
from hspylib.addons.widman.widgets.widget_time_calc import WidgetTimeCalc
from hspylib.core.enums.exit_status import ExitStatus
from hspylib.core.tools.commons import sysout
from hspylib.core.tools.zoned_datetime import now
from hspylib.modules.cli.application.argparse.argument_parser import HSArgumentParser
from hspylib.modules.cli.icons.font_awesome.widget_icons import WidgetIcons
from hspylib.modules.cli.vt100.terminal import Terminal


class WidgetPunch(Widget):
    """HSPyLib Widget to Report current system memory usage"""

    WIDGET_ICON = WidgetIcons.PUNCH
    WIDGET_NAME = "Punch"
    TOOLTIP = "!!! PUNCH THE CLOCK !!!"
    USAGE = dedent("""
    "Usage: ${FUNCNAME[0]} [options] <args>"

      Options: '
        -l        : List all registered punches.'
        -e        : Edit current punch file.'
        -r        : Reset punches for the current week and save the previous one.'
        -w <week> : Report (list) all punches of specified week using the pattern: week-N.punch.'

      Notes: '
        When no arguments are provided it will !!PUNCH THE CLOCK!!.'
    """)
    VERSION = (0, 1, 0)

    HHS_DIR = os.getenv('HHS_PUNCH_FILE', os.getenv('HOME', './'))

    HHS_PUNCH_FILE = os.getenv('HHS_PUNCH_FILE', f"-{HHS_DIR}/.punches")

    DATE_STAMP=now('%a %d-%m-%Y')

    TIME_STAMP=now('%H:%M')

    WEEK_STAMP=int(now('%V'))

    RE_TODAY_PUNCH_LINE = rf"({DATE_STAMP}).*"

    RE_PUNCH_LINE = r'^((Mon|Tue|Wed|Thu|Fri|Sat|Sun) )(([0-9]+-?)+) =>.*'

    MAX_PUNCHES = 7  # 7 week days

    @staticmethod
    def _daily_total(daily_punches: List[str], decimal: bool = False) -> str:
        """Calculate the total time from the daily punches."""
        # Up to 3 pairs of timestamps: morning, afternoon and evening
        if (n := len(daily_punches)) > 0 and n % 2 == 0:
            stamps = []
            stamps += [daily_punches[1], '-', daily_punches[0], '+'] if n // 2 >= 1 else []  # Morning
            stamps += [daily_punches[3], '-', daily_punches[2], '+'] if n // 2 >= 2 else []  # Afternoon
            stamps += [daily_punches[5], '-', daily_punches[4]] if n // 2 >= 3 else []       # Evening
            h, m, _ = WidgetTimeCalc.calc_time(stamps)
            m = WidgetTimeCalc.to_decimal(m) if decimal else m
            return f"{'%GREEN%'if h >= 8 else '%RED%'}{h:02d}{'.' if decimal else ':'}{m:02d}%NC%"
        else:
            return f"%RED%--{'.' if decimal else ':'}--%NC%"

    def __init__(self) -> None:
        super().__init__(
            self.WIDGET_ICON,
            self.WIDGET_NAME,
            self.TOOLTIP,
            self.USAGE,
            self.VERSION)

        self._exit_code = ExitStatus.SUCCESS
        self._fn = self._do_the_punch
        self._args = None
        self._today = None
        self._punches = []
        self._total_hour = self._total_min = 0
        self._week_num = self.WEEK_STAMP

    def execute(self, args: List[str] = None) -> ExitStatus:

        # Create the current week punch file if it does not yet exist.
        if not os.path.exists(self.HHS_PUNCH_FILE):
            with open(self.HHS_PUNCH_FILE, 'w') as f_punch:
                f_punch.write(f"{now('%d-%m-%Y')} => ")

        ret_val = self._parse_args(args)

        if not ret_val:
            return ExitStatus.ABORTED

        if self._args:
            if 'list' == self._args.action:
                self._punches = self._read_punches(self.HHS_PUNCH_FILE)
                self._fn = self._list_punches
            elif 'week' == self._args.action:
                punch_dir = os.path.dirname(self.HHS_PUNCH_FILE)
                self._week_num = self._args.week_num
                self._punches = self._read_punches(f"{punch_dir}/week-{self._week_num:02d}.punch")
                self._fn = self._list_punches
            elif 'edit' == self._args.action:
                self._fn = self._edit_punches
            elif 'reset' == self._args.action:
                self._fn = self._reset_punches
        else:
            self._punches = self._read_punches(self.HHS_PUNCH_FILE)

        self._fn()

        return ExitStatus.SUCCESS

    def _parse_args(self, args: List[str]) -> bool:
        """Parse command line arguments"""

        if not args:
            return True

        parser = HSArgumentParser(
            prog='punch', prefix_chars="+",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description='PUNCH-THE-CLOCK. This is a helper tool to aid with the timesheet.')

        subparsers = parser.add_subparsers(title='action', dest='action')
        subparsers.add_parser('list', help='list all registered punches.')
        subparsers.add_parser('edit', help='edit current punch file.')
        subparsers.add_parser('reset', help='reset punches for the current week and save the previous one.')

        w_parser = subparsers.add_parser('week', help='list all punches of the specified week-num (week-N.punch).')
        w_parser.add_argument('week_num', type=int, default=1, help='the week number')

        self._args = parser.parse_args(args)

        return bool(self._args)

    def _read_punches(self, punch_file: str) -> List[str]:
        """Read all punches from the punch file."""
        if not os.path.exists(punch_file):
            sysout(f"%RED%Punch file '{punch_file}' not found !%NC%")
            sys.exit(int(str(ExitStatus.FAILED.value)))
        else:
            with open(punch_file, 'r') as f_punch:
                all_punches = list(map(
                    self._set_today, filter(lambda l: re.match(self.RE_PUNCH_LINE, l), f_punch.readlines())
                ))
                if len(all_punches) > self.MAX_PUNCHES:
                    sysout(f"%RED%Punch file contains more than {self.MAX_PUNCHES} punch lines !%NC%")
                    sys.exit(int(str(ExitStatus.FAILED.value)))
                return all_punches

    def _is_today(self, punch_line: str) -> bool:
        """Whether the punch line refer to today's date."""
        return bool(re.match(self.RE_TODAY_PUNCH_LINE, punch_line))

    def _set_today(self, punch_line: str) -> str:
        """Set the today's date if the line represents today."""
        if not self._today and self._is_today(punch_line):
            self._today = punch_line
        return punch_line

    def _do_the_punch(self) -> None:
        """!!DO THE PUNCH!!"""
        with open(self.HHS_PUNCH_FILE, 'w') as f_punch:
            if not self._today:  # Write the first punch of the day
                self._today = f"{self.DATE_STAMP} => {self.TIME_STAMP}"
                self._punches.append(self._today)
            elif self._today.count(':') < 6:  # Only allowed 3 groups of 3 pairs of punches
                pat = rf"({self.DATE_STAMP}) => (.*)"
                if mat := re.match(pat, self._today):
                    self._today = re.sub(pat, f"{mat.group(1)} => {mat.group(2)} {self.TIME_STAMP} ", self._today)
                else:
                    raise Exception('Invalid punch file')
                self._punches[-1] = self._today
            list(map(f_punch.write, [f"{l.strip()}\n" for l in self._punches]))
        sysout(f"{re.sub(self.DATE_STAMP, '%GREEN%Today%NC%', self._today)} ")

    def _list_punches(self) -> None:
        """List all punches from the punch file."""
        total = 0, 0
        total_dec = 0, 0
        sysout(f"\n%WHITE%Week-{self._week_num:02d} Punches%NC%")
        sysout('-' * 82)
        for punch_line in self._punches:
            daily_punches = punch_line[17:].strip().split()
            n = len (daily_punches)
            padding = '.' * (35 if n == 0 else (36 - n * 6))
            line_color = '%BLUE%' if self._is_today(punch_line) else ''
            sysout(f"{line_color}{punch_line[:17]} {' '.join(daily_punches) + (' ' if n % 2 != 0 else '')}", end='')
            daily_total = self._daily_total(daily_punches)
            daily_total_dec = self._daily_total(daily_punches, decimal = True)
            if n > 0 and n % 2 == 0:
                sysout(f" {padding} : Subtotal = {daily_total} -> {daily_total_dec}%NC%")
                total = WidgetTimeCalc.calc_time([f"{total[0]}:{total[1]}", '+', daily_total[-9:-4]])
                total_dec = total[0], WidgetTimeCalc.to_decimal(total[1])
            else:
                sysout(f"{daily_total}%NC%")
        sysout('-' * 82)
        bh, bm, _ = WidgetTimeCalc.calc_time([f"{total[0]}:{total[1]}", '-', '40:00'])
        balance = f"{'%BLUE%' if bh >= 0 else '%RED%'}{bh:02d}:{bm:02d}%NC%"
        totals = f"{total[0]:02d}:{total[1]:02d} -> {total_dec[0]:02d}.{total_dec[1]:02d}"
        sysout(f"%WHITE%Total: ({totals})  Balance: {balance}\n")

    def _edit_punches(self) -> None:
        """Open the default system editor to edit punches."""
        Terminal.open(self.HHS_PUNCH_FILE)

    def _reset_punches(self) -> None:
        """Rename the punch file as a weekly punch file and reset current punch file."""
        punch_dir = os.path.dirname(self.HHS_PUNCH_FILE)
        new_name = f"{punch_dir}/week-{self._week_num:02d}.punch"
        if os.path.exists(new_name):
            sysout(f"%RED%Punch file '{new_name}' already exists!")
        else:
            os.rename(self.HHS_PUNCH_FILE, new_name)
            sysout(f"%YELLOW%Punch file {self.HHS_PUNCH_FILE} renamed to {new_name}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: hspylib
   @package: hspylib.main.hspylib.addons.appman
      @file: appman_enums.py
   @created: Fri, 29 Jul 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration

class Extension(Enumeration):
    """Appman available extensions"""
    GRADLE = 'gradle'
    GIT = 'git'


class AppType(Enumeration):
    """Appman supported application types"""
    APP = 'app'
    QT_APP = 'qt-app'
    WIDGET = 'widget'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.file
      @file: file_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
import contextlib
import logging as log
import sqlite3
from sqlite3 import Error
from typing import Iterable, List, Optional, Set, Tuple, TypeVar

from hspylib.core.datasource.crud_entity import CrudEntity
from hspylib.core.datasource.db_repository import Connection, Cursor, DBRepository, ResultSet, Session
from hspylib.core.datasource.identity import Identity
from hspylib.core.datasource.sqlite.sqlite_configuration import SQLiteConfiguration
from hspylib.core.exception.exceptions import DatabaseConnectionError, DatabaseError
from hspylib.core.tools.namespace import Namespace
from hspylib.core.tools.text_tools import quote

T = TypeVar('T', bound=CrudEntity)


class SQLiteRepository(DBRepository[T, SQLiteConfiguration]):
    """Implementation of a data access layer for a SQLite persistence store."""

    def __init__(self, config: SQLiteConfiguration):
        super().__init__(config)

    @property
    def db_file(self) -> str:
        return self._config.db_file

    @property
    def info(self) -> str:
        return f"{self.db_file}"

    def _create_session(self) -> Tuple[Connection, Cursor]:
        log.debug(f"{self.logname} Attempt to connect to database: {str(self)}")
        conn = sqlite3.connect(self.db_file)
        return conn, conn.cursor()

    @contextlib.contextmanager
    def _session(self) -> Session:
        log.debug(f"{self.logname} Attempt to connect to database: {str(self)}")
        conn, dbs = None, None
        try:
            conn, dbs = self._create_session()
            log.debug(f"{self.logname} Successfully connected to database: {self.info} [{self.db_file}]")
            yield dbs
        except Error as err:
            raise DatabaseConnectionError(f"Unable to open/execute-on database session => {err}") from err
        finally:
            if conn:
                log.debug(f"{self.logname} Closing connection [ssid={self.db_file}]")
                conn.commit()
                conn.close()

    def execute(self, sql_statement: str, **kwargs) -> Tuple[int, Optional[ResultSet]]:
        """TODO"""
        with self._session() as dbs:
            try:
                log.debug(f"{self.logname} Executing SQL statement {sql_statement} [ssid={hash(dbs)}]:\n"
                          f"\t|-Arguments: {str([f'{k}={v}' for k, v in kwargs.items()])}\n"
                          f"\t|-Statement: {sql_statement}")
                rows = dbs.execute(sql_statement, **kwargs).fetchall()
                return len(rows), rows
            except (sqlite3.ProgrammingError, sqlite3.OperationalError) as err:
                raise DatabaseError(f"Unable to execute statement => {sql_statement}") from err

    def count(self) -> int:
        sql = f"SELECT COUNT(*) FROM {self.table_name()}"
        return int(self.execute(sql)[1][0][0])

    def delete(self, entity: T) -> None:
        self.delete_by_id(entity.identity)

    def delete_by_id(self, entity_id: Identity) -> None:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"DELETE FROM {self.table_name()} WHERE " + ' AND '.join(clauses)
        self.execute(sql)

    def delete_all(self, entities: List[T]) -> None:
        values, s = [], entities[0]
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"DELETE FROM " \
              f"{self.table_name()} WHERE ({s.as_columns()}) IN ({', '.join(values)}) "
        self.execute(sql)

    def save(self, entity: T, exclude_update: Optional[Iterable[str]] = None) -> None:
        sql = f"INSERT INTO " \
              f"{self.table_name()} ({entity.as_columns()}) VALUES {entity.values} " \
              f"ON CONFLICT DO UPDATE SET {entity.as_column_set(prefix='EXCLUDED.', exclude=exclude_update)}"
        self.execute(sql)

    def save_all(self, entities: List[T], exclude_update: Optional[Iterable[str]] = None) -> None:
        values, s = [], entities[0]
        list(map(lambda e: values.append(str(e.values)), entities))
        sql = f"INSERT INTO " \
              f"{self.table_name()} ({s.as_columns()}) VALUES {', '.join(values)} " \
              f"ON CONFLICT DO UPDATE SET {s.as_column_set(prefix='EXCLUDED.', exclude=exclude_update)}"
        self.execute(sql)

    def find_all(
        self,
        fields: Optional[Set[str]] = None,
        filters: Optional[Namespace] = None,
        order_bys: Optional[List[str]] = None,
        limit: int = 500, offset: int = 0) -> List[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = list(filter(None, [f for f in filters.values])) if filters else None
        orders = list(filter(None, order_bys)) if order_bys else None
        sql = f"SELECT {fields} FROM {self.table_name()} " \
              f"{('WHERE ' + ' AND '.join(clauses)) if clauses else ''} " \
              f"{('ORDER BY ' + ', '.join(orders)) if orders else ''} " \
              f"LIMIT {limit} OFFSET {offset}"

        return list(map(self.to_entity_type, self.execute(sql)[1]))

    def find_by_id(
        self,
        entity_id: Identity,
        fields: Optional[Set[str]] = None) -> Optional[T]:

        fields = '*' if not fields else ', '.join(fields)
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT {fields} FROM {self.table_name()} " \
              f"WHERE {' AND '.join(clauses)}"
        result = next((e for e in self.execute(sql)[1]), None)

        return self.to_entity_type(result) if result else None

    def exists_by_id(self, entity_id: Identity) -> bool:
        clauses = [f"{k} = {quote(v)}" for k, v in zip(entity_id.attributes, entity_id.values)]
        sql = f"SELECT EXISTS(SELECT 1 FROM {self.table_name()} WHERE {' AND '.join(clauses)})"

        return self.execute(sql)[1][0][0] > 0
